import { assertValidName, astFromValue, Kind, } from 'graphql';
import { NexusWrappedSymbol } from '../core';
import { mapObj } from '../utils';
import { isNexusDirective, isNexusDirectiveUse } from './wrapping';
import { NexusTypes, withNexusSymbol } from './_types';
export const RequestDirectiveLocation = [
    /** Request Definitions */
    'QUERY',
    'MUTATION',
    'SUBSCRIPTION',
    'FIELD',
    'FRAGMENT_DEFINITION',
    'FRAGMENT_SPREAD',
    'INLINE_FRAGMENT',
    'VARIABLE_DEFINITION',
];
export const SchemaDirectiveLocation = [
    /** Type System Definitions */
    'SCHEMA',
    'SCALAR',
    'OBJECT',
    'FIELD_DEFINITION',
    'ARGUMENT_DEFINITION',
    'INTERFACE',
    'UNION',
    'ENUM',
    'ENUM_VALUE',
    'INPUT_OBJECT',
    'INPUT_FIELD_DEFINITION',
];
export class NexusDirectiveUse {
    constructor(name, args, config) {
        this.name = name;
        this.args = args;
        this.config = config;
    }
}
withNexusSymbol(NexusDirectiveUse, NexusTypes.DirectiveUse);
/** Defines a directive, which can be used positionally when generating the SDL */
export function directive(config) {
    assertValidName(config.name);
    config = Object.freeze(config);
    function addDirective(...args) {
        return new NexusDirectiveUse(config.name, args[0], config);
    }
    addDirective[NexusWrappedSymbol] = NexusTypes.Directive;
    addDirective.value = config;
    return addDirective;
}
export function addDirective(directiveName, ...args) {
    return new NexusDirectiveUse(directiveName, args[0]);
}
const DirectiveASTKindMapping = {
    SCALAR: Kind.SCALAR_TYPE_DEFINITION,
    SCHEMA: Kind.SCHEMA_DEFINITION,
    OBJECT: Kind.OBJECT_TYPE_DEFINITION,
    FIELD_DEFINITION: Kind.FIELD_DEFINITION,
    ARGUMENT_DEFINITION: Kind.INPUT_VALUE_DEFINITION,
    INTERFACE: Kind.INTERFACE_TYPE_DEFINITION,
    UNION: Kind.UNION_TYPE_DEFINITION,
    ENUM: Kind.ENUM_TYPE_DEFINITION,
    ENUM_VALUE: Kind.ENUM_VALUE_DEFINITION,
    INPUT_OBJECT: Kind.INPUT_OBJECT_TYPE_DEFINITION,
    INPUT_FIELD_DEFINITION: Kind.INPUT_VALUE_DEFINITION, // 'InputValueDefinition',
};
/**
 * Creates the ASTNode with the directives
 *
 * @param
 */
export function maybeAddDirectiveUses(kind, directiveUses, customDirectives) {
    if (!(directiveUses === null || directiveUses === void 0 ? void 0 : directiveUses.length)) {
        return undefined;
    }
    const seenDirectives = new Set();
    return {
        astNode: {
            kind: DirectiveASTKindMapping[kind],
            directives: Array.from(directiveUses).map((d) => {
                const directiveName = isNexusDirective(d) ? d.value.name : d.name;
                const directiveDef = customDirectives[directiveName];
                if (seenDirectives.has(directiveName) && !directiveDef.isRepeatable) {
                    throw new Error(`Cannot use directive ${directiveName} more than once in a row`);
                }
                assertValidDirectiveFor(kind, directiveDef);
                seenDirectives.add(directiveName);
                const args = isNexusDirectiveUse(d) ? directiveArgs(d, directiveDef) : undefined;
                return {
                    kind: Kind.DIRECTIVE,
                    name: {
                        kind: Kind.NAME,
                        value: directiveName,
                    },
                    arguments: args,
                };
            }),
        },
    };
}
function assertValidDirectiveFor(kind, directiveDef // any is a hack to make this work w/ v15 & 16
) {
    if (!directiveDef.locations.includes(kind)) {
        throw new Error(`Directive ${directiveDef.name} cannot be applied to ${kind}`);
    }
}
function directiveArgs(directiveUse, directiveDef) {
    var _a;
    return mapObj((_a = directiveUse.args) !== null && _a !== void 0 ? _a : {}, (val, key) => {
        const arg = directiveDef.args.find((a) => a.name === key);
        if (!arg) {
            throw new Error(`Unknown directive arg ${key}, expected one of ${directiveDef.args.map((a) => a.name)}`);
        }
        const value = astFromValue(val, arg.type);
        if (!value) {
            throw new Error(`Unable to get ast for ${key}`);
        }
        return {
            kind: Kind.ARGUMENT,
            name: {
                kind: Kind.NAME,
                value: key,
            },
            value,
        };
    });
}
//# sourceMappingURL=directive.js.map